//==========================================================================
//      QR-Faktorisierung einer Matrix nach HOUSEHOLDER
// Rueckkehrcode: Anzahl der gefundenen Nullspalten.
//==========================================================================
#include "ls.h"
int ls_qr(int m,       // Zeilenzahl der Matrix
          int n,       // Spaltenzahl der Matrix
          REAL *A,     // (m,n)-Matrix
                       // O: rechts der Haupdiagonalen ist das obere Dreieck
                       //    der oberen Dreiecksmatrix R abgelegt
                       //    (ohne Haupdiagonalelemente),
                       //    links der Haupdiagonalen (einschliesslich)
                       //    stehen die u-Vektoren fuer die  Matrix Q
          REAL **diag, // Hilfsfeld (oder NULL)
                       // O: Hauptdiagonalelemente von R
          REAL **q)    // Hilfsfeld (oder NULL)
                       // O: Faktoren fuer die u-Vektoren
{ int i,j,k,rc=0;
  REAL epsqr=1.e-10,*a,*aj,*ak,*ae=A+m*n,beta,s,rho,eins=1.0,
       *dia=*diag,*qq=*q;
  if(!dia) dia=*diag=new REAL[m];
  if(!qq) qq=*q=new REAL[m];
  for(k=0,a=A;a<ae;a+=n,k++)
  { for(ak=a+k,rho=0;ak<ae;rho+=*ak*(*ak),ak+=n);
    if(rho<epsqr){ dia[k]=0;rc++;continue;}
    rho=sqrt(rho);if(a[k]>0) rho=-rho;
    dia[k]=rho,a[k]-=rho,beta=-rho*a[k],qq[k]=beta=eins/beta;
    for(j=k+1;j<n;j++)
    { for(aj=a+j,ak=a+k,s=0;aj<ae;s+=*aj*(*ak),aj+=n,ak+=n);
      for(s*=beta,aj=a+j,ak=a+k;aj<ae;*aj-=*ak*s,aj+=n,ak+=n);}
  }
  return(rc);
}
