//=========================================================================
//            Funktionen der Klasse sp_bMatrix
//=========================================================================
#ifndef sp_h
#include "sp.h"
#endif
sp_bMatrix::sp_bMatrix(Int nn, Int d, char *meth, char *onam):sp_Org(onam)
{ static char *q1 = "b_Matrix::b_Matrix";
  n = nn, o_anz = d;
  if (!nn || !d) return;
  Int i;

  for (i = 0, typ = sp_END; sp_b_mask[i] != sp_END; i++)
    if (!strcmp(meth, sp_meth[sp_b_mask[i]])){ typ = sp_b_mask[i]; break;}

  switch(typ)
  { case sp_MAT:
      break;
    case sp_END:
      sp_error_(q1, 39);
  }

  if (!(o_ind = new Int [o_anz+1])) sp_error_(q1);
  memset(o_ind, 0, sizeof(Int)*(o_anz+1));
  if (!(o_f = new unsigned char [o_anz+1])) sp_error_(q1);
  memset(o_f, 0,o_anz+1);
  if (!(o_w = new Real* [o_anz+1])) sp_error_(q1);
  memset(o_w, 0, sizeof(Real*)*(o_anz+1));
}

sp_bMatrix::~sp_bMatrix()
{ if (!*copy && n)
  { Int i;
    for (i = 0; i <= o_anz; i++)
      if (o_w[i]) delete [] o_w[i];
    delete [] o_w, delete [] o_ind, delete [] o_f;
  }
}

void sp_bMatrix::write(ostream &f)
{ static char *q1 = "sp_bMatrix::write";
  Int i, prec = asp-9;
  Real *a, *ae;

  f.unsetf(ios::scientific), f.unsetf(ios::floatfield);

  if (oname) f << oname << endl;
  f<<"{ "<<"sp_bMatrix"<<" "<<n<<" "<<o_anz\
      << " "<< sp_meth[typ&&15] << endl;

  for (i = 0; i <= o_anz; f << endl, i++)
  { if (o_w[i])
    { f << "Diagonal " << o_ind[i] << endl;
      a = o_w[i], ae = (o_f[i])? a+1 : a+n-o_ind[i];
      for (; a < ae; a++)
           f << setw(asp) << setprecision(prec) << *a;
     }
  }
   f << endl << "}" << endl;
}

void sp_bMatrix::read(istream &f)
{ static char *q1 = "sp_bMatrix::read";
  char w[32];
  Real *a;
  Int nn, anz, ll, i, k;

  if (oname)
  { while (f >> w && strcmp(w, oname));
    if (!f.good()) sp_error_(q1, 9);
  }

  if (!(f >> w) || (*w != '{')) sp_error_(q1, 9);
  if (!(f >> w) || strcmp (w, "sp_bMatrix")) sp_error_(q1, 10);
  if (!(f >> nn)|| (n != nn) || !(cin >> anz) || (anz != o_anz)\
      || !(f >> w) || strcmp(w,sp_meth[typ>>4])) sp_error_(q1, 12);

  k = 0;
  while (f >> w)
  { if (*w == '}') break;
    if (strcmp (w, "Diagonal")) sp_error_(q1,10);
    f >> i;
    if ((k >= o_anz) || o_ind[k] ) sp_error_(q1,22);
    o_ind[k] = i, ll = n-i;
    if (!(a = o_w[k] = new Real [ll])) sp_error_(q1);
    while (ll--) f >> *a, a++;
    k++;
  }

  if (!f.good()) sp_error_(q1, 11);
}

void sp_bMatrix::operator>>(char *name)
{ if (*name == '\0') write(cout);
  else{ ofstream f(name); write(f);}
}

void sp_bMatrix::operator<<(char *name)
{ if (*name == '\0') read(cin);
  else{ ifstream f(name); read(f);}
}

void sp_bMatrix::operator<< (const Mat_el &u)          // NNE-Eingabe
{ static char * q1 = "sp_bMatrix::operator<<";
  if (!u.x) return;
  Int i = u.i, j = u.j, k;

  if (j < i) i = u.j, j = u.i;
  j -= i;

  for (k = 0; k <= o_anz; k++)
    if (o_ind[k] == j) { o_w[k][i-1] = u.x; return; }

  sp_error_(q1,22);
}

Real sp_bMatrix::operator>> (Mat_el *u)                // NNE-Ausgabe
{ static char * q1 = "sp_bMatrix::operator>>";
  Int i = u->i, j = u->j, k;

  if (j < i) i = u->j, j = u->i;
  j -= i;

  for (k = 0; k <= o_anz; k++)
    if (o_ind[k] == j) { u->x = o_w[k][i-1]; return u->x; }

  sp_error_(q1,22);
  return 0;
}

ls_Vector sp_bMatrix::operator* (ls_Vector &u) const
{ Int i, j, k, nn;
  ls_Vector v(n);
  Real s, *a, *uu = u.A, *vv = v.A;

  if (o_f[0])
    for (s = *o_w[0], i = 0; i < n; vv[i++] = s * uu[i]);
  else
    for (a = o_w[0], i = 0; i < n; vv[i++] = a[i] * uu[i]);

  for (k = 1; k <= o_anz; k++)
    if (o_f[k])
      for (s = *o_w[k], j = o_ind[k], nn = n-j, i = 0; i < nn;
         vv[i] += (s * uu[i+j]), vv[i+j] += (s * uu[i]), i++);
    else
      for (a = o_w[k], j = o_ind[k], nn = n-j, i = 0; i < nn;
         vv[i] += (a[i] * uu[i+j]), vv[i+j] += (a[i] * uu[i]), i++);

  return v;
}

void sp_bMatrix::put_upper_diagonal(Int idiag, const ls_Vector &u)
{ static char * q1 = "sp_bMatrix::put_upper_diagonal";
  Int i = idiag, ll = n-idiag;
  Real *a;

  if (idiag)
    for (i = 1; (i <= o_anz) && o_ind[i] && (o_ind[i] != idiag); i++);

  if ((i > o_anz) || o_w[i]) sp_error_(q1,22);
  if (!(a = o_w[i] = new Real [ll])) sp_error_(q1);
  memcpy(a, u.A,sizeof(Real)*ll), o_ind[i] = idiag;
}

void sp_bMatrix::put_upper_diagonal(Int idiag, Real s)
{ static char * q1 = "sp_bMatrix::put_upper_diagonal";
  Int i = idiag;
  Real *a;

  if (idiag)
    for ( i = 1; (i <= o_anz) && o_ind[i] && (o_ind[i] != idiag); i++);

  if ((i > o_anz) || o_w[i]) sp_error_(q1,22);
  if (!(a = o_w[i] = new Real)) sp_error_(q1);
  o_ind[i] = idiag, o_f[i] = 1, *a = s;
}

ls_Vector sp_bMatrix::get_upper_diagonal(Int idiag)
{ static char * q1 = "sp_bMatrix::get_upper_diagonal";
  Int i = idiag, ll = n-idiag;
  Real *a;

  if (idiag)
    for ( i = 1; (i <= o_anz) && (o_ind[i] != idiag); i++);

  if (i > o_anz) sp_error_(q1,22);
  if (!o_w[i]) sp_error_(q1,24);

  a = o_w[i];
  if (o_f[i]) ll = 1;
  ls_Vector u(ll);
  memcpy(u.A,a,sizeof(Real)*ll);
  return u;
}

void sp_bMatrix::delete_upper_diagonal(Int idiag)
{ static char *q1 = "sp_bMatrix::delete_upper_diagonal";
  Int i = idiag, ll = n-idiag;

  if (idiag)
    for ( i = 1; (i <= o_anz) && (o_ind[i] != idiag); i++);
  if (i > o_anz) sp_error_(q1,22);
  if (o_w[i]) delete []o_w[i],o_w[i] = NULL, o_ind[i]=0,o_f[i]=0;
}

ls_Vector sp_bMatrix::solve_with_MAT(ls_Vector &x, ls_Vector &b)
{ ls_Vector p(n), w(n);
  b -= *this * x, p = b;
  Real a, bb = b * b, bbn, wp;
  Int it = 5 * n, jt = it; *rc = 0;

  if (bb < eps) return x;

  while (jt--)
  { w = *this * p, wp = w * p;
    if (fabs(wp) < eps) { *rc = it - jt; return x; }
    a = bb/wp, x += a * p; b -= a * w, bbn = b * b,
    p = b + (bbn/bb) * p, bb = bbn;
  }

  if (bb < eps) { *rc = it - jt; return x; }
  b -= *this * x, p = *this * b, bb = b * b, wp = p * p;
  jt = it;

  while (jt--)
  { a = bb/wp; x += a * p; b -= a * (*this * p), bbn = b * b,
    p = (*this * b) + (bbn/bb) * p, wp = p * p, bb = bbn;
    if (wp < eps) { *rc = it + it - jt; return x; }
  }

  return x;
}

ls_Vector sp_bMatrix::solve(ls_Vector &b)
{ ls_Vector x(n); 
  return solve_with_MAT(x, b); 
}
