//==========================================================================
//              Fehlerroutine und Restfunktionen
//==========================================================================
#ifndef sp_h
#include "sp.h"
#endif
void sp_error_(const char *v, Int l)
{ static char *a[] = {
    "no memory.",                   // 0
    "missing Vector name.",         // 1
    "cannot open Vector file.",     // 2
    "Vector not found.",            // 3
    "Vector end error.",            // 4
    "no correct Vector.",           // 5
    "Vector size error.",           // 6
    "Vector component index error.",// 7
    "cannot open Matrix file.",     // 8
    "Matrix not found.",            // 9
    "no correct Matrix.",           // 10
    "Matrix contents error.",       // 11
    "Matrix size error.",           // 12
    "operation/function forbidden.",// 13
    "index error.",                 // 14
    "row index error.",             // 15
    "column index error.",          // 16
    "Matrix singular.",             // 17
    "s_Matrix singular.",           // 18
    "Matrix write error.",          // 19
    "Matrix read error.",           // 20
    "incomplete s_Matrix singular.",// 21
    "diagonal index error.",        // 22
    "missing diagonals.",           // 23
    "diagonal is empty.",           // 24
    "missing file name.",           // 25
    "matrix method error.",         // 26
    "matrix is not empty.",         // 27
    "method not found.",            // 28
    "incomplete q_Matrix singular." // 29
  };
  cerr << "*** " << l <<" "<< v << ": "; cerr << a[l] << endl;
  exit(1);
}


sp_Org::sp_Org(char *onam)
{ asp = (sizeof(Real) < 8)? dr4 : dr8,
  eps = (sizeof(Real) < 8)? eps4 : eps8;
  copy = new Int, *copy = 0,
  rc = new Int, *rc = 0, oname = onam;
}

sp_Org::sp_Org(const sp_Org &B)
{ asp = B.asp, eps = B.eps, copy = B.copy, *copy += 1, rc = B.rc;}

sp_Org::~sp_Org()
{ if (*copy) { *copy -= 1; return; }
  delete rc, delete copy;
}

void sp_Org::sp_zero_(NNE **A, Int *l_A, Int m, Int n)
{ static char *q1 = "sp_Org::sp_zero_";
  NNE *a, *aa, *ae;
  Int mm;
  if (n)
  { if (!(a = *A = new NNE[m+1])) sp_error_(q1);
    a->i = m+1, a->j = n+1;
    for (aa = a+1, ae = a+m+1; aa < ae;
          aa->z = aa->s = a, aa->x = 0, aa->i = aa->j = 0, aa++);
    if (!(a = new NNE[More+1])) sp_error_(q1);
    a->s = NULL, (*A)->z = a+1, (*A)->s = a;
    for (aa = a+1, ae = a+More; aa < ae; aa->z = aa+1, aa++);
    ae->z = NULL, *l_A = More-1;
    return;
  }
  mm = m/More; if (!m || (mm*More < m)) mm++;
  while (mm--)
  { if (!(a = new NNE[More+1])) sp_error_(q1);
    for (aa = a+1, ae = a+More; aa < ae; aa->z = aa+1, aa++);
    ae->z = NULL, a->s = (*A)->s, (*A)->s = a,
    ((a->s)+More)->z = a+1, (a->s)->z = a, *l_A += More;
  }
}

void sp_Org::sp_delete_(NNE *A)
{ NNE *a, *aa;
  for(a = A; a; aa = a->s, delete [] a, a = aa);
}

ls_Org::ls_Org (char *onam)
{ asp = (sizeof(Real) < 8)? dr4 : dr8,
  eps = (sizeof(Real) < 8)? eps4 : eps8;
  copy = new Int, *copy = 0;
  oname = onam;
}

ls_Org::ls_Org (const ls_Org &B)
{ asp = B.asp, eps = B.eps, copy = B.copy, *copy += 1, oname = B.oname; }

ls_Org::~ls_Org()
{ if (!*copy) delete copy; else *copy -= 1; }

Real ausl(Real a, Real b)  // Ausloeschung zweier Zahlen
{ if (!a || !b || (a * b < 0)) return fabs(a-b);
  a = fabs(a), b = fabs(b);
  if ((a < 1) && (b < 1)) return fabs(a-b);
  return fabs(a - b)/((a > b)? a : b);
}

ls_Vector operator* (Real a, const ls_Vector &x)
{ Int n = x.n, i;
  ls_Vector vv(n); vv = x;
  Real *v = vv.A;
  for (i = 0; i < n; v[i++] *= a);
  return vv;
}