//=========================================================================
//            Funktionen der Klasse sp_Matrix
//=========================================================================
#ifndef sp_h
#include "sp.h"
#endif
sp_Matrix::sp_Matrix(Int mm, Int nn, char *meth, char *onam):sp_Org(onam)
{ static char *q1 = "sp_Matrix::sp_Matrix";
  n = nn, m = mm, l_A = 0;
  if (!mm || !nn) return;
  Int i;

#ifdef sp_test
  if (m < n) sp_error_(q1,14);
#endif

  for (i = 0, typ = sp_END; sp_m_mask[i] != sp_END; i++)
    if (!strcmp(meth, sp_meth[sp_m_mask[i]])){ typ = sp_m_mask[i]; break;}

  switch(typ)
  { case sp_MAT:
      sp_zero_(&A, &l_A, m, n);
      typ <<= 4;
      break;
    case sp_END:
      sp_error_(q1, 39);
  }
}

sp_Matrix::~sp_Matrix()
{ if (!*copy && n)
  { switch(typ>>4)
    { case sp_MAT:
        sp_delete_(A);
        break;
    }
  }
}

void sp_Matrix::write(ostream &f)
{ static char *q1 = "sp_Matrix::write";
  Int i, prec = asp-9;
  NNE *iz;

  f.unsetf(ios::scientific), f.unsetf(ios::floatfield);

  if (oname) f << oname << endl;
  f << "{ sp_Matrix" << m << " " << n << " " << sp_meth[typ&15] << endl;

  switch(typ&15)
  { case sp_MAT:
      for (i = 1; i <= m; i++)
      { f << "Row " << i << " " << (A+i)->i << " " << (A+i)->j << endl;
        iz = A+i;
        while ((iz = iz->z) != A)
          f << setw(10) << iz->j\
               << setw(asp) << setprecision(prec) << iz->x << endl;
      }
      break;
    default:
      sp_error_(q1, 26);
  }
  f << "}" << endl;
}

void sp_Matrix::read(istream &f)
{ static char *q1 = "sp_Matrix::read";
  if (typ&15) sp_error_(q1, 13);
  char w[32];
  unsigned char atyp;
  Mat_el u;
  Int nn, mm, ll, i, k;

  if (oname)
  { while (f >> w && strcmp(w, oname));
    if (!f.good()) sp_error_(q1, 9);
  }
  if (!(f >> w) || (*w != '{')) sp_error_(q1, 9);
  if (!(f >> w) || strcmp (w, "sp_Matrix")) sp_error_(q1, 10);
  if (!(f >> mm) || !(cin >> nn)\
      || (n != nn) || (m != mm)) sp_error_(q1, 12);
  if (!(f >> w)) sp_error_(q1, 10);
   for (i = 0, atyp = sp_END; sp_m_mask[i] != sp_END; i++)
    if (!strcmp(w, sp_meth[sp_m_mask[i]])){ atyp = sp_m_mask[i]; break;}

  switch(atyp)
  { case sp_MAT:
      for (k = 1; k <= m; k++)
      { if ((A+k)->i) sp_error_(q1, 27);
        f >> w;
        if (strcmp(w, "Row")) sp_error_(q1, 11);
        f >> i;
        if (i != k) sp_error_(q1, 15);
        f >> (A+i)->i, f >> (A+i)->j, u.i=i, ll = (A+i)->i;
        while (ll--) f >> u.j, f >> u.x, *this << u;
      }
      break;
    default:
      sp_error_(q1,26);
  }
  if (!(f >> w) || (*w != '}')) sp_error_(q1, 10);
}

void sp_Matrix::operator>>(char *name)
{ if (*name == '\0') write(cout);
  else{ ofstream f(name); write(f);}
}

void sp_Matrix::operator<<(char *name)
{ if (*name == '\0') read(cin);
  else{ ifstream f(name); read(f);}
}

Real sp_Matrix::filling_density()
{ Int i, ll = 0, ne = m*n;
  if (get_name())
    for (i = 0; i < m; ll += (A+i)->i, i++);
  else
    for (i = 0; i < m; ll += (A+i)->i+(A+i)->i, i++);
  return Real((ll+0.)/ne);
}

void sp_Matrix::operator<< (const Mat_el &mat)
{
#ifdef sp_test
  static char *q1 = "sp_Matrix::operator<<";
  if (typ&15) sp_error_(q1, 13);
#endif

  Int j, i;
  Real s;
  NNE *js, *iz, *l;
  i = mat.i, j = mat.j, s = mat.x;
  if (!s) return;

  if (!l_A) sp_zero_(&A, &l_A);
  js = A+j; while (js->s->i < i) js = js->s;
  if (js->s->i == i) { js->s->x = s; return; }
  iz = A+i, iz->i += 1, (A+j)->j += 1 ;
  while (iz->z->j < j) iz = iz->z;
  l = A->z, A->z = l->z, l->x = s, l->z = iz->z, l->s = js->s,
  l->i = i, l->j = j, js->s = iz->z = l, l_A--;
}

Real sp_Matrix::operator>> (Mat_el *mat)
{ Int j = mat->j, i = mat->i;
  NNE *iz = A+i; while ((iz = iz->z)->j < j);
  if (iz->j == j) mat->x = iz->x; else mat->x = 0;
  return mat->x;
}

sp_Vector sp_Matrix::operator* (sp_Vector &vv) const
{
#ifdef sp_test
  static char *q1 = "sp_Matrix::operator*";
  if (vv.n != n) sp_error_(q1, 6);
  if (typ&15) sp_error_(q1, 13);
#endif
  Int i, k, ll, le = vv.n_A;
  Vec_el *v, b;
  NNE *iz;
  sp_Vector u(m);
  Real s;

  for (i = 1; i <= m; i++)
  { s = 0, iz = A+i, k = 0, v = vv.A;
    while ((k < le) && (ll = (iz = iz->z)->j) <= n)
    { while ((k < le) && (v->i < ll)) k++, v++;
      if (v->i == ll) s += iz->x * v->x, v++;
    }
    if (!s) continue;
    b.i = i, b.x = s, u << b;
  }
  return u;
}

ls_Vector sp_Matrix::operator* (ls_Vector &v) const
{
#ifdef sp_test
  static char *q1 = "sp_Matrix::operator*";
  if (v.n != n) sp_error_(q1, 6);
  if (typ&15) sp_error_(q1, 13);
#endif
  Int i, ll;
  Real s;
  NNE *iz;
  ls_Vector u(m);

  for (i = 1; i <= m; u[i++] = s)
    for (s=0,iz=A+i;(ll = (iz = iz->z)->j)<=n;s+=iz->x * v[ll]);
  return u;
}

ls_Vector sp_Matrix::operator% (ls_Vector &v) const
{
#ifdef sp_test
  static char *q1 = "sp_Matrix::operator%";
  if (v.n != m) sp_error_(q1, 6);
  if (typ&15) sp_error_(q1, 13);
#endif
  Int j, ll;
  NNE *js;
  Real s;
  ls_Vector u(n);

  for (j = 1; j <= n; u[j++] = s)
    for (s=0,js=A+j;(ll = (js = js->s)->i)<=m;s+=js->x * v[ll]);
  return u;
}

void sp_Matrix::print()
{ Int i, j, k, l, mm, nn;
  NNE *iz;
  mm = (m > 8)?  8 : m;
  nn = (n > 8)?  8 : n;

  for ( i = 1; i <= mm; cout << endl, i++)
  {  iz = A+i, j = 1, k = 0;
     while ((iz = iz->z) != A)
     { if ((k = iz->j) > nn) k = nn;
       for (l = j; l < k; l++) cout << setw(10) << setprecision(4) << 0.;
       if (k != iz->j) break;
       cout << setw(10) << setprecision(4) << iz->x;
       j = k+1;
     }
     for (l = k+1; l <= nn; l++) cout << setw(10) << 0.;
  }
  cout << endl;
}

sp_Vector sp_Matrix::get_row(Int i)
{
#ifdef sp_test
  static char *q1 = "sp_Matrix::get_row";
  if (i > m) sp_error_(q1, 15);
#endif
  NNE *js, *iz;
  int na = get_name();
  Int ll = (!na)? (A+i)->i + (A+i)->j : (A+i)->i;
  sp_Vector vv(n,ll+3);
  Vec_el *v = vv.A;

  if (!na)
  { js = A+i;
    while ((js = js->s)->i < i) v->i = js->i, v->x = js->x, v++;
  }
  iz = A+i;
  while ((iz = iz->z) != A) v->i = iz->j, v->x = iz->x, v++;
  vv.n_A = ll;
  return vv;
}

void sp_Matrix::put_row(Int i, const sp_Vector &vv)
{ static char *q1 = "sp_Matrix::put_row";
#ifdef sp_test
  if (i > m) sp_error_(q1, 15);
  if (typ&15) sp_error_(q1, 13);
#endif
  Int ll = vv.n_A, j, k;
  NNE *js, *iz, *l;
  int na = get_name();
  Real s;
  Vec_el *v = vv.A;

  if (l_A < ll) sp_zero_(&A, &l_A, ll-l_A);

  while (ll--)
  { if (!(s = v->x)) { v++; continue; }
#ifdef sp_test
    if ((j = v->i) > n) sp_error_(q1, 16);
#endif
    if (!na && (j < i)) k = j, j = i; else k = i;
    v++, js = A+j; while (js->s->i < k) js = js->s;
    if (js->s->i == k) { js->s->x = s; continue; }
    iz = A+k, iz->i += 1, (A+j)->j += 1;
    while (iz->z->j < j) iz = iz->z;
    l = A->z, A->z = A->z->z, l->x = s, l->z = iz->z, l->s = js->s,
    l->i = k, l->j = j, js->s = iz->z = l, l_A--;
  }
}

sp_Vector sp_Matrix::get_column(Int j)
{ if (!get_name()) return get_row(j);
#ifdef sp_test
  static char *q1 = "sp_Matrix::get_column";
  if (j > n) sp_error_(q1, 16);
#endif
  NNE *js = A+j;
  Int ll = (A+j)->j;
  sp_Vector vv(n, ll+3);
  Vec_el *v = vv.A;

  while ((js = js->s) != A) v->i = js->i, v->x = js->x, v++;
  vv.n_A = ll;
  return vv;
}

void sp_Matrix::put_column(Int j, const sp_Vector &vv)
{ if (!get_name()) { put_row(j, vv); return; }
  static char *q1 = "sp_Matrix::put_column";
#ifdef sp_test
  if (j > n) sp_error_(q1, 16);
  if (typ&15) sp_error_(q1, 13);
#endif
  Int ll = vv.n_A, i;
  NNE *js, *iz, *l;
  Real s;
  Vec_el *v = vv.A;

  if (l_A < ll) sp_zero_(&A, &l_A, ll - l_A);

  while (ll--)
  { if (!(s = v->x)) { v++; continue; }
#ifdef sp_test
    if ((i = v->i) > m) sp_error_(q1, 16);
#endif
    v++, js = A+j; while (js->s->i < i) js = js->s;
    if (js->s->i == i) { js->s->x = s; continue; }
    iz = A+i, iz->i += 1, (A+j)->j += 1;
    while (iz->z->j < j) iz = iz->z;
    l = A->z, A->z = A->z->z, l->x = s, l->z = iz->z, l->s = js->s,
    l->i = i, l->j = j, js->s = iz->z = l, l_A--;
  }
}

sp_Vector sp_Matrix::get_lower_diagonal(Int idiag)
{ if (!get_name()) { return get_upper_diagonal(idiag); }
#ifdef sp_test
  static char *q1 = "sp_Matrix::get_lower_diagonal";
  if (idiag > m) sp_error_(q1, 15);
#endif
  Int i = idiag+1, j = 1, l = 1;
  NNE *iz;

  while ((i <= m) && (j <= n))
  { iz = A+i; while ((iz = iz->z)->j < j);
    if (iz->j == j) l++;
    i++, j++;
  }

  sp_Vector vv(m-idiag, l);
  Vec_el *v = vv.A;

  i = idiag+1, j = 1;
  while ((i <= m) && (j <= n))
  { iz = A+i; while ((iz = iz->z)->j < j);
    if (iz->j == j) v->i = j, v->x = iz->x, v++;
    i++, j++;
  }
  vv.n_A = l-1;
  return vv;
}

void sp_Matrix::put_lower_diagonal(Int idiag, const sp_Vector &vv)
{ if (!get_name()) { put_upper_diagonal(idiag,vv); return; }
  static char *q1 = "sp_Matrix::put_lower_diagonal";
#ifdef sp_test
  if (idiag > m) sp_error_(q1, 15);
  if (typ&15) sp_error_(q1, 13);
#endif
  Int ll = vv.n_A, i, j;
  NNE *js, *iz, *l;
  Real s;
  Vec_el *v = vv.A;

  if (l_A < ll) sp_zero_(&A, &l_A, ll - l_A);

  while (ll--)
  { if (!(s = v->x)) { v++; continue; }
#ifdef sp_test
    if ((j = v->i) > n) sp_error_(q1, 16);
#endif
    if ((i = j+idiag) > m) break;
    v++, js = A+j; while (js->s->i < i) js = js->s;
    if (js->s->i == i) { js->s->x = s; continue; }
    iz = A+i, iz->i += 1, (A+j)->j += 1;
    while (iz->z->j < j) iz = iz->z;
    l = A->z, A->z = A->z->z, l->x = s, l->z = iz->z, l->s = js->s,
    l->i = i, l->j = j, js->s = iz->z = l, l_A--;
  }
}

void sp_Matrix::put_lower_diagonal(Int idiag, Real s)
{ if (!get_name()) { put_upper_diagonal(idiag,s); return; }
#ifdef sp_test
  static char *q1 = "sp_Matrix::put_lower_diagonal";
  if (idiag > m) sp_error_(q1, 15);
  if (typ&15) sp_error_(q1, 13);
#endif
  if (fabs(s) < eps) return;
  NNE *js, *iz, *l;
  Int i, j, ll = n-idiag;

  if (l_A < ll) sp_zero_(&A, &l_A, ll-l_A);

  for (j = 1; j <= ll; j++)
  { i = j+idiag;
    js = A+j; while (js->s->i < i) js = js->s;
    if (js->s->i == i) { js->s->x = s; continue; }
    iz = A+i, iz->i += 1, (A+j)->j += 1;
    while (iz->z->j < j) iz = iz->z;
    l = A->z, A->z = A->z->z, l->x = s, l->z = iz->z, l->s = js->s,
    l->i = i, l->j = j, js->s = iz->z = l, l_A--;
  }
}

sp_Vector sp_Matrix::get_upper_diagonal(Int idiag)
{
#ifdef sp_test
  static char *q1 = "sp_Matrix::get_upper_diagonal";
  if (idiag > n) sp_error_(q1, 16);
#endif
  Int ll = 1, j = idiag+1, i = 1;
  NNE *iz;

  while ((i <= m) && (j <= n))
  { iz = A+i; while ((iz = iz->z)->j < j);
    if (iz->j == j) ll++;
    i++, j++;
  }

  sp_Vector vv(n-idiag, ll);
  Vec_el *v = vv.A;

  j = idiag+1, i = 1;
  while ((i <= m) && (j <= n))
  { iz = A+i; while ((iz = iz->z)->j < j);
    if (iz->j == j) v->i = i, v->x = iz->x, v++;
    i++, j++;
  }
  vv.n_A = ll-1;
  return vv;
}

void sp_Matrix::put_upper_diagonal(Int idiag, const sp_Vector &vv)
{ static char *q1 = "sp_Matrix::put_upper_diagonal";
#ifdef sp_test
  if (idiag > n) sp_error_(q1, 16);
  if (typ&15) sp_error_(q1, 13);
#endif
  Int ll = vv.n_A, i, j;
  NNE *js, *iz, *l;
  Real s;
  Vec_el *v = vv.A;

  if (l_A < ll) sp_zero_(&A, &l_A, ll-l_A);

  while (ll--)
  { if (!(s = v->x)) { v++; continue; }
#ifdef sp_test
    if ((i = v->i) > m) sp_error_(q1, 16);
#endif
    if ((j = i+idiag) > n) break;
    v++, js = A+j; while (js->s->i < i) js = js->s;
    if (js->s->i == i) { js->s->x = s; continue; }
    iz = A+i, iz->i += 1, (A+j)->j += 1;
    while (iz->z->j < j) iz = iz->z;
    l = A->z, A->z = A->z->z, l->z = iz->z, l->s = js->s, l->x = s,
    l->i = i, l->j = j, js->s = iz->z = l, l_A--;
  }
}

void sp_Matrix::put_upper_diagonal(Int idiag, Real s)
{
#ifdef sp_test
  static char *q1 = "sp_Matrix::put_upper_diagonal";
  if (idiag > n) sp_error_(q1, 16);
  if (typ&15) sp_error_(q1, 13);
#endif
  Int ll = n-idiag, i, j;
  NNE *js, *iz, *l;

  if(fabs(s) < eps) return;

  if (l_A < ll) sp_zero_(&A, &l_A, ll-l_A);

  for (i = 1; i <= ll; i++)
  { j = i+idiag;
    js = A+j; while (js->s->i < i) js = js->s;
    if (js->s->i == i) { js->s->x = s; continue; }
    iz = A+i, iz->i += 1, (A+j)->j += 1;
    while (iz->z->j < j) iz = iz->z;
    l = A->z, A->z = A->z->z, l->x = s, l->z = iz->z, l->s = js->s,
    l->i = i, l->j = j, js->s = iz->z = l, l_A--;
  }
}

void sp_Matrix::delete_nne(const Mat_el &mat)
{ static char *q1 = "sp_Matrix::delete_nne";
  Int i, j, k;
  NNE  *js, *iz, *l;
  int na = get_name();

  i = mat.i, j = mat.j;
  if (!na && (j < i)) k = j, j = i; else k = i;
#ifdef sp_test
  if ((k > m) || (j > n)) sp_error_(q1, 14);
  if (typ&15) sp_error_(q1, 13);
#endif
  js = A+j; while (js->s->i < k) js = js->s;
  if (js->s->i != k) return;
  iz = A+k, iz->i -= 1, (A+j)->j -= 1;
  while (iz->z != js->s) iz = iz->z;
  l = A->z, A->z = iz->z, iz->z = iz->z->z, A->z->z = l,
  js->s = js->s->s, l_A++;
}

void sp_Matrix::delete_row(Int i)
{
#ifdef sp_test
  static char *q1 ="sp_Matrix::delete_row";
  if (typ&15) sp_error_(q1, 13);
#endif
  NNE *js, *iz, *l;

  iz = A+i;
  while ((iz = iz->z) != A)
  { js = A+(iz->j), js->j -= 1;
    while (js->s != iz) js = js->s;
    js->s = js->s->s, js = iz;
  }

  if ((A+i)->i)
    l = A->z, A->z = (A+i)->z, js->z = l, (A+i)->z = A,
    l_A += (A+i)->i, (A+i)->i = 0;

  if (get_name()) return;

  js = A+i;
  while ((js = js->s) != A)
  { iz = A+(js->i), iz->i -= 1;
    while (iz->z != js) iz = iz->z;
      l = A->z, A->z = js, iz->z = iz->z->z, A->z->z = l;
  }
  (A+i)->s = A, l_A += (A+i)->j, (A+i)->j = 0;
}

void sp_Matrix::delete_in_row(Int i, const sp_Vector &vv)
{
#ifdef sp_test
  static char *q1 = "sp_Matrix::delete_in_row";
  if (vv.n != n) sp_error_(q1, 6);
  if (typ&15) sp_error_(q1, 13);
#endif
  Int j, k, le = vv.n_A;
  NNE *js, *iz, *l;
  Vec_el *v = vv.A;
  int na = get_name();

  while (le--)
  { j = v->i, v++;
    if (!na && (j < i)) k = j, j = i; else k = i;
    js = A+j; while (js->s->i < i) js = js->s;
    if (js->s->i != i) continue;
    iz = A+i, (A+j)->j -= 1, iz->i -= 1;
    while (iz->z != js->s) iz = iz->z;
    l = A->z, A->z = iz->z, iz->z = iz->z->z, A->z->z = l,
    js->s = js->s->s, l_A++;
  }
}

void sp_Matrix::delete_column(Int j)
{
#ifdef sp_test
  static char *q1 ="sp_Matrix::delete_column";
  if (typ&15) sp_error_(q1, 13);
#endif
  if (!get_name()) { delete_row(j); return; }
  NNE *js, *iz, *l;

  js = A+j;
  while ((js = js->s) != A)
  { iz = A+(js->i), iz->i -= 1;
    while (iz->z != js) iz = iz->z;
    l = A->z, A->z = js, iz->z = iz->z->z, A->z->z = l;
  }
   (A+j)->s = A, l_A += (A+j)->j, (A+j)->j = 0;
}

void sp_Matrix::delete_in_column(Int j, const sp_Vector &vv)
{ if (!get_name()) { delete_in_row(j, vv); return; }
#ifdef sp_test
  static char *q1 = "sp_Matrix::delete_in_column";
  if (vv.n != n) sp_error_(q1, 6);
  if (typ&15) sp_error_(q1, 13);
#endif
  Vec_el *v = vv.A;
  Int i, le = vv.n_A;
  NNE *js, *iz, *l;

  while (le--)
  { i = v->i, v++, js = A+j; while (js->s->i < i) js = js->s;
    if (js->s->i != i) continue;
    iz = A+i, iz->i -= 1, (A+j)->j -= 1;
    while (iz->z != js->s) iz = iz->z;
    l = A->z, A->z = iz->z, iz->z = iz->z->z, A->z->z = l,
    js->s = js->s->s, l_A++;
  }
}

void sp_Matrix::delete_lower_diagonal(Int idiag)
{
#ifdef sp_test
  static char *q1 = "sp_Matrix::delete_lower_diagonal";
  if (typ&15) sp_error_(q1, 13);
#endif
  if (!get_name()) { delete_upper_diagonal(idiag); return; }
  Int j, i, le;
  NNE *js, *iz, *l;

  le = (m-idiag > n)? n : m-idiag;
  j = 0;
  while (le--)
  { j++, i = j+idiag;
    js = A+j; while (js->s->i < i) js = js->s;
    if (js->s->i != i) continue;
    iz = A+i, iz->i -= 1, (A+j)->j -= 1;
    while (iz->z != js->s) iz = iz->z;
    l = A->z, A->z = iz->z, iz->z = iz->z->z, A->z->z = l,
    js->s = js->s->s, l_A++;
  }
}

void sp_Matrix::delete_in_lower_diagonal(Int idiag, const sp_Vector &vv)
{
#ifdef sp_test
  static char *q1 = "sp_Matrix::delete_in_lower_diagonal";
  if (typ&15) sp_error_(q1, 13);
#endif
  if (!get_name()) { delete_in_upper_diagonal(idiag, vv); return; }
  Vec_el *v = vv.A;
  Int j, i, le = vv.n_A;
  NNE *js, *iz, *l;

  while (le--)
  { j = v->i, i = j+idiag, v++;
    js = A+j; while (js->s->i < i) js = js->s;
    if (js->s->i != i) continue;
    iz = A+i, iz->i -= 1, (A+j)->j -= 1;
    while (iz->z != js->s) iz = iz->z;
    l = A->z, A->z = iz->z, iz->z = iz->z->z, A->z->z = l,
    js->s = js->s->s, l_A++;
  }
}

void sp_Matrix::delete_upper_diagonal(Int idiag)
{
#ifdef sp_test
  static char *q1 = "sp_Matrix::delete_upper_diagonal";
  if (typ&15) sp_error_(q1, 13);
#endif
  Int j, i, le;
  NNE *js, *iz, *l;

  le = (n-idiag > n)? n : n-idiag;
  i = 0;
  while (le--)
  { i++, j = i+idiag, iz = A+i; while (iz->z->j < j) iz = iz->z;
    if (iz->z->j != j) continue;
    js = A+j, js->j -= 1, (A+i)->i -= 1;
    while (js->s != iz->z) js = js->s;
    l = A->z, A->z = iz->z, iz->z = iz->z->z, A->z->z = l,
    js->s = js->s->s, l_A++;
  }
}

void sp_Matrix::delete_in_upper_diagonal(Int idiag, const sp_Vector &vv)
{
#ifdef sp_test
  static char *q1 = "sp_Matrix::delete_upper_diagonal";
  if (typ&15) sp_error_(q1, 13);
#endif
  Vec_el *v = vv.A;
  Int j, i, le = vv.n_A;
  NNE *js, *iz, *l;

  while (le--)
  { i = v->i, j = i+idiag, v++, iz = A+i; while (iz->z->j < j) iz = iz->z;
    if (iz->z->j != j) continue;
    js = A+j, js->j -= 1, (A+i)->i -= 1;
    while (js->s != iz->z) js = js->s;
    l = A->z, A->z = iz->z, iz->z = iz->z->z, A->z->z = l,
    js->s = js->s->s, l_A++;
  }
}

void sp_Matrix::change_row(Int i, Int k) // i<k
{
#ifdef sp_test
  static char *q1 = "sp_Matrix::change_row";
  if ((i >= k) || (i > m) || (k > m)) sp_error_(q1, 14);
  if (typ&15) sp_error_(q1, 13);
#endif
  sp_Vector ai(n), ak(n);
  ai = get_row(i), ak = get_row(k);
  delete_row(i), delete_row(k);
  put_row(i,ak),put_row(k,ai);
}

void sp_Matrix::change_column(Int j, Int k) // j<k
{
#ifdef sp_test
  static char *q1 = "sp_Matrix::change_column";
  if ((j >= k) || (j > n) || (k > n)) sp_error_(q1, 14);
  if (typ&15) sp_error_(q1, 13);
#endif
  sp_Vector aj(n), ak(n);
  aj = get_column(j), ak = get_column(k);
  delete_column(j), delete_column(k);
  put_column(j,ak),put_column(k,aj);
}

ls_Vector sp_Matrix::solve_with_MAT(ls_Vector &x, ls_Vector &b)
{
#ifdef sp_test
  static char *q1 = "sp_Matrix::solve_with_MAT";
  if (typ&15) sp_error_(q1, 13);
#endif
  ls_Vector p(n);

  b -= *this * x, p = *this % b;
  Real s, bb = b * b, bbn, pp = p * p;
  Int it = 5*m, jt = it; *rc = 0;
  if (pp < eps) { *rc = it-jt; return x; }

  while (jt--)
  { s = bb/pp, x += s * p, b -= s * (*this * p), bbn = b * b,
    p = (*this % b) + (bbn/bb) * p, pp = p * p, bb = bbn;
    if (pp < eps) break;
  }
  *rc = it-jt;
  return x;
}

ls_Vector sp_Matrix::solve(ls_Vector &b)
{ static char *q1 = "sp_Matrix::solve";
  ls_Vector x(n);

  switch(typ&15)
  { case sp_MAT: return solve_with_MAT(x, b);
    default:
      sp_error_(q1,26);
  }
  return x;
}
