//=========================================================================
//            Funktionen der Klasse sp_qMatrix
//=========================================================================
#ifndef sp_h
#include "sp.h"
#endif
sp_qMatrix::sp_qMatrix(Int nn, char *meth, \
                       char *onam):sp_Matrix(0,0,meth,onam)
{ static char *q1 = "sp_qMatrix::sp_qMatrix";
  Int i;
  m = n = nn, l_A = 0, irow = icol = NULL;
  if (!n) return; 

  for (i = 0, typ = sp_END; sp_q_mask[i] != sp_END; i++)
    if (!strcmp(meth, sp_meth[sp_q_mask[i]])){ typ = sp_q_mask[i]; break;}

  switch(typ)
  { case sp_MAT:
    case sp_LU:
    case sp_LUinc:
      sp_zero_(&A, &l_A, m, n);
      typ <<= 4;
      break;
    case sp_LUpiv:
      sp_zero_(&A, &l_A, m, n);
      if (!(irow = new Int [n+1])) sp_error_(q1);
      if (!(icol = new Int [n+1])) sp_error_(q1);
      typ <<= 4;
      break;
    case sp_END:
      sp_error_(q1, 39);
  }
}

sp_qMatrix::~sp_qMatrix()
{ if (!*copy && n)
  { switch(typ>>4)
    { case sp_MAT:
      case sp_LU:
        delete []A;
        break;
      case sp_LUinc:
        if (a_name && !a_n){ remove(a_name); delete [] a_name;}
        if (u_name && !u_n){ remove(u_name); delete [] u_name;;}
        delete []A;
        break;
      case sp_LUpiv:
        delete []A;
        if (irow) delete [] irow, delete [] icol;
        break;
    }
    typ = (sp_END<<4);
  }
}

void sp_qMatrix::write(ostream &f)
{ static char *q1 = "sp_qMatrix::write";
  Int i, prec = asp-9;
  NNE *iz;

  f.unsetf(ios::scientific), f.unsetf(ios::floatfield);

  if (oname) f << oname << endl;
  f << "{ sp_qMatrix" << " " << n;

  switch (typ&15)
  { case sp_MAT:
      f << " " << sp_meth[typ&15] << endl;
      for (i = 1; i <= n; i++)
      { f << "Row " << i << " " << (A+i)->i << " " << (A+i)->j << endl;
        iz = A+i;
        while ((iz = iz->z) != A)
          f << setw(10) << iz->j\
            << setw(asp) << setprecision(prec) << iz->x << endl;
      }
      break;
    case sp_LU:
      f << " " << sp_meth[typ&15] << endl;
      for (i = 1; i <= n; i++)
      { f <<"Row "<<i<<" "<<(A+i)->i<<" "<<(A+i)->j<<endl;
        iz = A+i;
        while ((iz = iz->z) != A)
          f << setw(10) << iz->j\
            << setw(asp) << setprecision(prec) << iz->x << endl;
      }
      break;
    case sp_LUpiv:
      f << " " << sp_meth[typ&15] << endl;
      for (i = 1; i <= n; f << " " << irow[i++]); f << endl;
      for (i = 1; i <= n; f << " " << icol[i++]); f << endl;
      for (i = 1; i <= m; i++)
      { f <<"Row "<<i<<" "<<(A+i)->i<<" "<<(A+i)->j<<endl;
        iz = A+i;
        while ((iz = iz->z) != A)
          f << setw(10) << iz->j\
            << setw(asp) << setprecision(prec) << iz->x << endl;
      }
      break;
    case sp_LUinc:
      if (!u_n) f << " " << sp_meth[sp_MAT] << endl; 
      else f << " " << sp_meth[typ&15] << endl;
      for (i = 1; i <= n; i++)
      { f <<"Row "<<i<<" "<<(A+i)->i<<" "<<(A+i)->j<<endl;
        iz = A+i;
        while ((iz = iz->z) != A)
          f << setw(10) << iz->j\
            << setw(asp) << setprecision(prec) << iz->x << endl;
      }
      if (u_n)
      { ifstream fu(u_name, ios::in|ios::binary);
        if (!fu) sp_error_(q1, 8);
        NNE *a;
        Int lm = sizeof(NNE)*(More+1), lmm = sizeof(NNE)*(n+1);
        for (a=A,fu.read((char *)a,lmm);a = a->s;fu.read((char *)a,lm));
        for (i = 1; i <= n; f << endl, i++)
        { iz = A+i;
          while ((iz = iz->z) != A)
            f << setw(asp) << setprecision(prec) << iz->x;
        }
        ifstream fa(a_name, ios::in|ios::binary);
        if (!fa) sp_error_(q1, 8);
        for (a=A,fa.read((char *)a,lmm);a = a->s;fa.read((char *)a,lm));
      }
      break;
  }
  f << "}" << endl;
}

void sp_qMatrix::read(istream &f)
{ static char *q1 = "sp_qMatrix::read";
  char w[32];
  unsigned char atyp;
  Mat_el u;
  Int nn, ll, i, k;

  if (oname)
  { while (f >> w && strcmp(w, oname));
    if (!f.good()) sp_error_(q1, 9);
  }
  if (!(f >> w) || (*w != '{')) sp_error_(q1, 9);
  if (!(f >> w) || strcmp (w, "sp_qMatrix")) sp_error_(q1, 10);
  if (!(f >> nn) || (n != nn)) sp_error_(q1, 12);
  if (!(f >> w)) sp_error_(q1, 10);
   for (i = 0, atyp = sp_END; sp_q_mask[i] != sp_END; i++)
    if (!strcmp(w,sp_meth[sp_q_mask[i]])){ atyp=sp_q_mask[i];break;}

  switch(atyp)
  { case sp_MAT:
    case sp_LU:
      if ((typ&15) != atyp) sp_error_(q1,26);
      for (k = 1; k <= n; k++)
      { if ((A+k)->i) sp_error_(q1, 27);
        f >> w;
        if (strcmp(w, "Row")) sp_error_(q1, 11);
        f >> i;
        if (i != k) sp_error_(q1, 15);
        f >> (A+i)->i, f >> (A+i)->j, u.i=i, ll = (A+i)->i;
        while (ll--) f >> u.j, f >> u.x, *this << u;
      }
      typ |= atyp;
      break;
    case sp_LUpiv:
      if ((typ>>4) != atyp) sp_error_(q1,26);
      for (i = 1; i <= n; f >> irow[i++]);
      for (i = 1; i <= n; f >> icol[i++]);
      for (k = 1; k <= n; k++)
      { if ((A+k)->i) sp_error_(q1, 27);
        f >> w;
        if (strcmp(w, "Row")) sp_error_(q1, 11);
        f >> i;
        if (i != k) sp_error_(q1, 15);
        f >> (A+i)->i, f >> (A+i)->j, u.i=i, ll = (A+i)->i;
        while (ll--) f >> u.j, f >> u.x, *this << u;
      }
      typ |= sp_LUpiv;
      break;
    case sp_LUinc:
      if (!(((typ>>4)==atyp)||((typ>>4)==sp_MAT)))sp_error_(q1,26);
      for (k = 1; k <= n; k++)
      { if ((A+k)->i) sp_error_(q1, 27);
        f >> w;
        if (strcmp(w, "Row")) sp_error_(q1, 11);
        f >> i;
        if (i != k) sp_error_(q1, 15);
        f >> (A+i)->i, f >> (A+i)->j, u.i=i, ll = (A+i)->i;
        while (ll--) f >> u.j, f >> u.x, *this << u;
      }
      if ((typ>>4) != atyp) break;
      if (!a_name)
        a_name = new char [L_tmpnam], tmpnam(a_name),
        u_name = new char [L_tmpnam], tmpnam(u_name);
      { NNE *a, *iz;
        Int lm = sizeof(NNE)*(More+1), lmm = sizeof(NNE)*(n+1);
        ofstream fa(a_name, ios::out|ios::binary);
        if (!fa) sp_error_(q1, 8);
        for(a=A,fa.write((char *)a,lmm);a=a->s;fa.write((char *)a,lm));
        fa.close();
        for (i = 1; i <= n; i++)
        { iz = A+i; while ((iz = iz->z) != A) f >> iz->x;}
        ofstream fu(u_name, ios::out|ios::binary);
        if (!fu) sp_error_(q1, 8);
        for(a=A,fu.write((char *)a,lmm);a=a->s;fu.write((char *)a,lm));
        ifstream ff(a_name, ios::in|ios::binary);
        if (!ff) sp_error_(q1, 8);
        for(a=A,ff.read((char *)a,lmm);a=a->s;ff.read((char *)a,lm));
      }
      typ |= sp_LUinc;
      break;
    default:
      sp_error_(q1, 26);
  }
  if (!(f >> w) || (*w != '}')) sp_error_(q1, 11);
}

void sp_qMatrix::operator>>(char *name)
{ if (*name == '\0') write(cout);
  else{ ofstream f(name); write(f);}
}

void sp_qMatrix::operator<<(char *name)
{ if (*name == '\0') read(cin);
  else{ ifstream f(name); read(f);}
}

void sp_qMatrix::operator<< (const Mat_el &mat)
{ Int j, i;
  Real s;
  NNE *js, *iz, *l;
  i = mat.i, j = mat.j, s = mat.x;
  if (!s) return;

  if (!l_A) sp_zero_(&A, &l_A);

  js = A+j; while (js->s->i < i) js = js->s;
  if (js->s->i == i) { js->s->x = s; return; }
  iz = A+i, iz->i += 1, (A+j)->j += 1 ;
  while (iz->z->j < j) iz = iz->z;
  l = A->z, A->z = l->z, l->x = s, l->z = iz->z, l->s = js->s,
  l->i = i, l->j = j, js->s = iz->z = l, l_A--;
}

Real sp_qMatrix::operator>> (Mat_el *mat)
{ Int j = mat->j, i = mat->i;
  NNE *iz = A+i; while ((iz = iz->z)->j < j);
  if (iz->j == j) mat->x = iz->x; else mat->x = 0;
  return mat->x;
}

ls_Vector sp_qMatrix::solve_with_LU(ls_Vector &b)
{ static char *q1 = "sp_qMatrix solve_with_LU";
  Int i, j, k, ll;
  Real s, ss;
  NNE *js, *ij, *iz, *idz, *ivs, *l;
  if ((typ>>4) != sp_LU) sp_error_(q1,26);

// LU-Faktorisierung


 if(!(typ&15))
  { for (i = 1; i <= n; (A+i)->x = s, i++)
      for (s = 0, iz = A+i; (iz = iz->z) != A; s += fabs(iz->x));
    for (i = 1; i <= n; i++)
    { idz = A+i; while ((idz = idz->z)->j < i);
      if ((idz->j != i) || fabs (idz->x) < eps) sp_error_(q1, 17);
      s = 1/idz->x, js = A+i; while ((js = js->s)->i < i);
      if (idz->z == A)
      { while ((js = js->s) != A) js->x *= s; continue; }
      ivs = js;
      while ((js = ivs->s) != A)
      { if (fabs (ss = js->x * s) < eps)
        { iz = A+(js->i), iz->i -= 1, (A+(js->j))->j -= 1;
          while (iz->z != js) iz = iz->z;
          l = A->z, A->z = js, iz->z = iz->z->z, A->z->z = l,
          ivs->s = ivs->s->s, l_A++;
          continue;
        }
        js->x = ss, ivs = js, ij = idz, k = js->i, iz = A+k;
        while ((j = (ij = ij->z)->j) <= n)
        { while (iz->z->j < j) iz = iz->z;
          if (iz->z->j == j) { iz->z->x -= ss * ij->x; continue; }
          if (!l_A) sp_zero_(&A, &l_A);
          js = A+j, js->j += 1, (A+k)->i += 1;
          while (js->s->i < k) js = js->s;
          l = A->z, A->z = A->z->z, l->x = -ss * ij->x, l->z = iz->z,
          l->s = js->s, l->i = k, l->j = j, js->s = iz->z = l, l_A--;
        }
      }
    }
    typ |= sp_LU;
  }
  if (!b.n) return b;
#ifdef sp_test
  if (b.n != n) sp_error_(q1, 6);
#endif
  ls_Vector x(n);
  Real *xx = x.A, *bb = b.A;

// loesen

  for (i = 1; i <= n; bb[i-1] -= s, i++)
    for (s = 0, iz = A+i; (ll = (iz = iz->z)->j) < i;
         s += iz->x * bb[ll-1]);
  for (i = n; i > 0; xx[i-1]= (bb[i-1]-s)/ss, i--)
  { iz = A+i; while ((iz = iz->z)->j < i);
    for (s = 0, ss = iz->x; (ll = (iz = iz->z)->j) <= n;
         s += iz->x * xx[ll-1]);
  }
  return x;
}

ls_Vector sp_qMatrix::solve_with_LUpiv(ls_Vector &b)
{ static char *q1 = "sp_qMatrix::solve_with_LUpiv";
  Int i, j, k, ll;
  Real s, ss, t;
  NNE *js, *ij, *iz, *idz, *ivs, *l;

  if ((typ>>4) != sp_LUpiv) sp_error_(q1,26);

  if (!(typ&15))
  {

// LUpiv-Faktorisierung mit Pivotisierung

    for (i = 1; i <= n; (A+i)->x = s, irow[i] = icol[i] = i, i++)
    { s = 0, iz = A+i; while ((iz = iz->z) != A) s += fabs(iz->x); }
    for (i = 1; i <= n; i++)
    { for (k = i, ll = (A+i)->i, j = i+1; j <= n; j++)
        if ((A+j)->i < ll) k = j, ll = (A+j)->i;
      if (i < k)
      {  change_row(i,k),
         ll = irow[i], irow[i] = irow[k], irow[k] = ll;
      }
      t = (A+i)->x, idz = A+i; while ((idz = idz->z)->j < i);
      s = (idz->j == i)? fabs(idz->x) * t : 0;
      if (s < eps)
      { iz = idz;
        while ((iz = iz->z) != A)
          if ((ss = fabs(iz->x) * t) > s) s = ss, idz = iz;
        if (s < eps) sp_error_(q1, 17);
        k = idz->j;
        if (i < k)
        { change_column(i,k),
          ll = icol[i], icol[i] = icol[k], icol[k] = ll;
          idz = A+i; while ((idz = idz->z)->j < i);
        }
      }
      s = 1/idz->x, js = A+i; while ((js = js->s)->i < i);
      if (idz->z == A) { while ((js = js->s) != A) js->x *= s; continue; }
      ivs = js;
      while ((js = ivs->s) != A)
      { if (fabs (ss = js->x * s) < eps)
        { iz = A+js->i, iz->i -= 1, (A+js->j)->j -= 1;
          while (iz->z != js) iz = iz->z;
          l = A->z, A->z = js, iz->z = iz->z->z, A->z->z = l,
          ivs->s = ivs->s->s, l_A++;
          continue;
        }
        js->x = ss, ivs = js, k = js->i, ij = idz, iz = A+k;
        while ((j = (ij = ij->z)->j) <= n)
        { while (iz->z->j < j) iz = iz->z;
          if (iz->z->j == j) { iz->z->x -= ss * ij->x; continue; }
          if (!l_A) sp_zero_(&A, &l_A);
          js = A+j, js->j += 1, (A+k)->i += 1;
          while (js->s->i < k) js = js->s;
          l = A->z, A->z = A->z->z, l->x = -ss * ij->x, l->z = iz->z,
          l->s = js->s, l->i = k, l->j = j, js->s = iz->z = l, l_A--;
        }
      }
    }
    typ |= sp_LUpiv;
  }

  if (!b.n) return b;
#ifdef sp_test
  if (b.n != n) sp_error_(q1, 6);
#endif

  ls_Vector x(n);
  Real *xx = x.A, *bb = b.A;

// loesen

  for (i = 1; i <= n; bb[irow[i]-1] -= s, i++)
    for (s = 0, iz = A+i; (ll = (iz = iz->z)->j) < i;
         s += iz->x * bb[irow[ll]-1]);
  for (i = n; i > 0; xx[icol[i]-1]= (bb[irow[i]-1]-s)/ss, i--)
  { iz = A+i; while ((iz = iz->z)->j < i);
    for (s = 0, ss = iz->x; (ll = (iz = iz->z)->j) <= n;
         s += iz->x * xx[icol[ll]-1]);
  }
  return x;
}

ls_Vector sp_qMatrix::solve_with_LUinc(ls_Vector &x, ls_Vector &b)
{ static char *q1 = "sp_qMatrix::solve_with_LUpre";
  Int i, j, ll;
  NNE *a, *ij, *idz, *iz, *js;
  Real s, ss;
  Int lm = sizeof(NNE)*(More+1), lmm = sizeof(NNE)*(n+1);
  if ((typ>>4) != sp_LUinc) sp_error_(q1, 26);

  if (!(typ&15))
  { if (!a_name) a_name = new char [L_tmpnam], tmpnam(a_name);else a_n=1;
    if (!u_name) u_name = new char [L_tmpnam], tmpnam(u_name);else u_n=1;
    { ofstream fa(a_name, ios::out|ios::binary);
      if (!fa) sp_error_(q1, 8);
      for (a=A,fa.write((char *)a,lmm);a=a->s;fa.write((char *)a,lm));
    }

// unvollstaendige LU-Zerlegung;  Elemente <eps werden genullt

    for (i = 1; i <= n; (A+i)->x = s, i++)
      for (s = 0, iz = A+i; (iz = iz->z) != A; s += fabs(iz->x));
    for (i = 1; i <= n; i++)
    { idz = A+i; while ((idz = idz->z)->j < i);
      s = (idz->j == i)? fabs(idz->x) * (A+i)->x : 0;
      iz = idz;
      if (s < eps) sp_error_(q1, 29);
      s = 1/idz->x, js = A+i; while ((js = js->s)->i < i);
      if (idz->z == A) { while ((js = js->s) != A) js->x *= s; continue; }
      while ((js = js->s) != A)
      { if (fabs(ss = js->x * s) < eps) { js->x = 0; continue; }
        js->x = ss, ij = idz, iz = A+js->i;
        while ((j = (ij = ij->z)->j) <= n)
        { while (iz->j < j) iz = iz->z;
          if (iz->j == j) iz->x -= ss * ij->x;
        }
      }
    }
    ofstream fu(u_name, ios::out|ios::binary);
    if (!fu) sp_error_(q1, 8);
    for (a=A,fu.write((char *)a,lmm);a=a->s;fu.write((char *)a,lm));
    ifstream fa(a_name, ios::in|ios::binary);
    if (!fa) sp_error_(q1, 8);
    for (a=A,fa.read((char *)a,lmm);a=a->s;fa.read((char *)a,lm));
    typ |= sp_LUinc;
  }

// vorkonditioniertes cg_Verfahren

  if (!b.n) return b;
#ifdef sp_test
  if (b.n != n) sp_error_(q1, 6);
#endif

  ifstream fa(a_name, ios::in|ios::binary);
  if (!fa) sp_error_(q1, 8);
  ifstream fu(u_name, ios::in|ios::binary);
  if (!fu) sp_error_(q1, 8);
  ls_Vector q(n), y(n), p(n);

  b -= *this * x;
  Real *bb=b.A, *qq=q.A, *yy=y.A, *pp=p.A;
  for (a=A,fu.read((char *)a,lmm);a = a->s;fu.read((char *)a,lm));
  fu.seekg(0, ios::beg);
  for (i = 1; i <= n; yy[i-1] = bb[i-1] - s, i++)
    for (s=0,iz=A+i;(ll = (iz = iz->z)->j)<i;s+=iz->x * yy[ll-1]);
  for (i = n; i > 0; qq[i-1] = (yy[i-1]-s)/ss, i--)
  { iz = A+i; while ((iz = iz->z)->j < i);
    for (s=0,ss=iz->x;(ll = (iz = iz->z)->j)<=n;s+=iz->x * qq[ll-1]);
  }
  Int it = 5*n, jt = it; *rc = 0;
  p = q;
  Real alpha, bq1 = b * q, bq2;

  while (jt--)
  { for (a=A,fa.read((char *)a,lmm);a = a->s;fa.read((char *)a,lm));
    fa.seekg(0, ios::beg);
    y = *this * p;
    if (fabs(s = p*y) < eps) break;
    alpha = bq1/s, x += alpha * p, b -= alpha * y;
    for (a=A,fu.read((char *)a,lmm);a=a->s;fu.read((char *)a,lm));
    fu.seekg(0, ios::beg);
    for (i = 1; i <= n; yy[i-1] = bb[i-1] - s, i++)
      for (s=0,iz=A+i;(ll=(iz = iz->z)->j)<i;s+=iz->x * yy[ll-1]);
    for (i = n; i > 0; qq[i-1] = (yy[i-1]-s)/ss,i--)
    { iz = A+i; while ((iz = iz->z)->j < i);
      for (s=0,ss=iz->x;(ll=(iz = iz->z)->j)<=n;s+=iz->x * qq[ll-1]);
    }
    bq2 = b * q, p = q + (bq2/bq1) * p, bq1 = bq2;
  }
  *rc = it-jt;
  return x;
}

ls_Vector sp_qMatrix::solve(ls_Vector &b)
{ static char *q1 ="sp_qMatrix::solve";
  ls_Vector x(n);

  switch(typ>>4)
  { case   sp_MAT: return solve_with_MAT(x, b);
    case    sp_LU: return solve_with_LU(b);
    case sp_LUpiv: return solve_with_LUpiv(b);
    case sp_LUinc: return solve_with_LUinc(x, b);
    default: sp_error_(q1, 26);
  }
  return x;
}
