//=========================================================================
//            Funktionen der Klasse sp_Vector
//=========================================================================
#ifndef sp_h
#include "sp.h"
#endif
sp_Vector::sp_Vector(Int nn, Int ll, char *onam):sp_Org(onam)
{ n = nn;
  if (!nn) return;
  if (ll) l_A = ll; else l_A =  (More > n)? n : More;
  A = new Vec_el[l_A];
  memset(A, 0, sizeof(Vec_el) * l_A);
  n_A = 0;
}

sp_Vector::~sp_Vector()
{ if (!*copy && n) delete [] A; }

void sp_Vector::write(ostream &f)
{ static char *q1 = "sp_Vector::write";
  Int prec = asp - 9;
  Vec_el *v, *ve;

  f.unsetf(ios::scientific), f.unsetf(ios::floatfield);

  if (oname) f << oname << endl;
  f << "{ sp_Vector " << n << " " << n_A << endl;

  for (v = A, ve = v+n_A; v < ve;
    f<<setw(10)<<v->i<<" "<<setw(asp)<<setprecision(prec)<<v->x<<endl,v++);
  f << "}" << endl;
}

void sp_Vector::read(istream &f)
{ static char *q1 = "sp_Vector::read";
  char w[32];
  Int nn, lle;
  Vec_el *v, *ve;
 
  if (oname)
  { while (f >> w && strcmp(oname, w));
    if (!f.good ()) sp_error_(q1, 3);
  }

  if (!(f >> w) || (*w != '{')) sp_error_(q1, 3);
  if (!(f >> w) || strcmp(w, "sp_Vector")) sp_error_(q1, 5);
  if (!(f >> nn) ||!(f >> lle) || (n != nn)) sp_error_(q1, 6);
  if (lle > n_A)
  { delete [] A; n_A = lle, l_A = (n_A+More > n)? n : n_A+More;
    if (!(A = new Vec_el[l_A])) sp_error_(q1);
    memset(A, 0, sizeof(Vec_el) * l_A);
  }
  n_A = lle;
  for (v = A, ve = v+n_A; v < ve; f >> v->i, f >> v->x, v++);
  if (!(f >> w) || (*w != '}')) sp_error_(q1, 4);
}

void sp_Vector::operator>>(char *name)
{ if (*name == '\0') write(cout);
  else{ ofstream f(name); write(f);}
}

void sp_Vector::operator<<(char *name)
{ if (*name == '\0') read(cin);
  else{ ifstream f(name); read(f);}
}

void sp_Vector::operator<< (const Vec_el &uu)// aufsteigende Komponenten
{ char *q1 = "sp_Vector::operator<<";
#ifdef sp_test
  if (uu.i > n) sp_error_(q1, 7);
#endif
  Int i = uu.i;
  Vec_el *a;

  if (n_A&& (uu.i < A[n_A-1].i)) sp_error_(q1, 7);

  if (n_A < l_A)
  { A[n_A].i = uu.i, A[n_A].x = uu.x, n_A++;
    return;
  }
  if (n_A == l_A)
  { l_A += More;
    if (!(a = new Vec_el[l_A])) sp_error_(q1);
    memset(a, 0, sizeof(Vec_el) * l_A);
    memcpy(a, A, sizeof(Vec_el) * n_A);
    delete [] A; A = a;
  }
  if (uu.i > A[n_A-1].i)
  { A[n_A].i = uu.i, A[n_A].x = uu.x, n_A++;  return; }
  Vec_el *v = A, *ve = v+n_A;
  while (v->i < i) v++;
  if (v->i > i) sp_error_(q1, 7);
  v->x = uu.x;
}

Real sp_Vector::operator[] (Int k)           // Komponenten-Ausgabe
{
#ifdef sp_test
  static char *q1 = "sp_Vector::operator[]";
  if (!(k > 0) || (k > n)) sp_error_(q1, 7);
#endif
  Vec_el *v = A, *ve = v+n_A;

  while ((v < ve) && (v->i < k)) v++;
  if (v->i == k) return v->x;
  return 0;
}

Int sp_Vector::operator== (const sp_Vector &u) const// Test auf Gleichheit
{
#ifdef sp_test
  static char *q1 = "sp_Vector::operator==";
  if (u.n != n) sp_error_(q1, 6);
#endif
  if (n_A != u.n_A) return 0;

  for (Vec_el *x = A, *xe = x+n_A, *y = u.A; x < xe; x++, y++)
  { if (x->i != y->i) return 0;
    if (ausl(x->x, -y->x) > eps) return 0;
  }
  return 1;
}

Int sp_Vector::operator!= (const sp_Vector &u) const// Test auf Ungleichheit
{
#ifdef sp_test
  static char *q1 = "sp_Vector::operator!=";
  if (u.n != n) sp_error_(q1, 6);
#endif
  if (n_A != u.n_A) return 1;

  for (Vec_el *x = A, *xe = x+n_A, *y = u.A; x < xe; x++, y++)
  { if (x->i != y->i) return 1;
    if (ausl(x->x, -y->x) > eps) return 1;
  }
  return 0;
}

sp_Vector sp_Vector::operator- () const             // negativer sp_Vector
{ sp_Vector yy(n); yy = *this;
  for (Vec_el *y = yy.A, *ye = y+n_A; y < ye; y->x = -y->x, y++);
  return yy;
}

sp_Vector sp_Vector::operator* (Real a) const       // sp_Vector mal Wert
{ sp_Vector yy(n); yy = *this;
  for (Vec_el *y = yy.A, *ye = y+n_A; y < ye; y->x *= a, y++);
  return yy;
}

void sp_Vector::operator= (const sp_Vector &u)  // sp_Vector = sp_Vector
{ static char *q1 = "sp_Vector::operator=";
  if (n) delete [] A;
  n_A = u.n_A, l_A = u.l_A;

  if (!(A = new Vec_el[l_A])) sp_error_(q1);

  memcpy(A, u.A, sizeof(Vec_el) * n_A);
}

void sp_Vector::operator= (const ls_Vector &u)  // sp_Vector = ls_Vector
{ static char *q1 = "sp_Vector::operator=";
  Int i;
  Real *f = u.A;

  if (n) delete [] A; n_A = 0;
  n = u.n;
  for (i = 0; i < n; i++) if (f[i]) n_A++;
  l_A = n_A;
  Vec_el *v;
  if (!(v = A = new Vec_el[l_A])) sp_error_(q1);
  for (i = 0; i < n; i++) if (f[i]) v->i = i, v->x = f[i];
}

sp_Vector sp_Vector::operator+ (const sp_Vector &uu) const
{
#ifdef sp_test
  static char *q1 = "sp_Vector::operator+";
  if (uu.n != n) sp_error_(q1, 6);
#endif
  sp_Vector yy(n, n_A + uu.n_A);
  Int yle = 0;
  Vec_el *y = yy.A, *u = uu.A, *v = A, *ue = u+uu.n_A, *ve = v+n_A;

  while ((u < ue) && (v < ve))
  { if (u->i == v->i)
    { y->i = u->i, y->x = v->x + u->x, u++, v++, yle++; continue; }
    if (u->i < v->i) { *y = *u, u++, y++, yle++; continue; }
    *y=*v, v++, y++, yle++;
  }
  if (u < ue) memcpy(y, u, sizeof(Vec_el) * (ue-u)), yle += (ue-u);
  if (v < ve) memcpy(y, v, sizeof(Vec_el) * (ve-v)), yle += (ve-v);
  yy.n_A = yle;
  return yy;
}

sp_Vector sp_Vector::operator- (const sp_Vector &uu) const
{
#ifdef sp_test
  static char *q1 = "sp_Vector::operator-";
  if (uu.n != n) sp_error_(q1, 6);
#endif
  sp_Vector yy(n, n_A + uu.n_A);
  Int yle = 0;
  Vec_el *y = yy.A, *u = uu.A, *v = A, *ue = u+uu.n_A, *ve = v+n_A;

  while ((u < ue) && (v < ve))
  { if (u->i == v->i)
    { y->i = u->i, y->x = v->x - u->x, u++, v++, yle++; continue; }
    if (u->i < v->i) { *y = *u, u++, y++, yle++; continue; }
    *y=*v, v++, y++, yle++;
  }

  while (u < ue) y->x = -u->x, y->i = u->i, yle++, u++;
  if (v < ve) memcpy(y, v, sizeof(Vec_el) * (ve-v)), yle += (ve-v);
  yy.n_A = yle; return yy;
}

void sp_Vector::operator+= (const sp_Vector &uu) // sp_Vector += sp_Vector
{
#ifdef sp_test
  static char *q1 = "sp_Vector::operator+=";
  if (uu.n != n) sp_error_(q1, 6);
#endif
  sp_Vector yy(n, n_A+uu.n_A);
  Int yle = 0;
  Vec_el *y = yy.A, *u = uu.A, *v = A, *ue = u+uu.n_A, *ve = v+n_A;

  while ((u < ue) && (v < ve))
  { if (u->i == v->i)
    { y->i = u->i, y->x = v->x + u->x, u++, v++, yle++; continue; }
    if (u->i < v->i) { *y = *u, u++, y++, yle++; continue; }
    *y=*v, v++, y++, yle++;
  }

  if (u < ue) memcpy(y, u, sizeof(Vec_el) * (ue-u)), yle += (ue-u);
  if (v < ve) memcpy(y, v, sizeof(Vec_el) * (ve-v)), yle += (ve-v);
  yy.n_A = yle; *this = yy;
}

void sp_Vector::operator-= (const sp_Vector &uu) // sp_Vector -= sp_Vector
{
#ifdef sp_test
  static char *q1 = "sp_Vector::operator-=";
  if (uu.n != n) sp_error_(q1, 6);
#endif
  sp_Vector yy (n, n_A + uu.n_A);
  Int yle = 0;
  Vec_el *y = yy.A, *u = uu.A, *v = A, *ue = u+uu.n_A, *ve = v+n_A;

  while ((u < ue) && (v < ve))
  { if (u->i == v->i)
    { y->i = u->i, y->x = v->x - u->x, u++, v++, yle++; continue; }
    if (u->i < v->i) { *y = *u, u++, y++, yle++; continue; }
    *y = *v, v++, y++, yle++;
  }

  while (u < ue) y->x = -u->x, y->i = u->i, yle++, u++;
  if (v < ve) memcpy(y, v, sizeof(Vec_el) * (ve-v)), yle += (ve-v);
  yy.n_A = yle; *this = yy;
}

Real sp_Vector::operator* (const sp_Vector &uu) const// Skalarprodukt
{
#ifdef sp_test
  static char *q1 = "sp_Vector::operator*";
  if (uu.n != n) sp_error_(q1, 6);
#endif
  Vec_el *u = uu.A, *ue = u+uu.n_A, *v = A, *ve = v+n_A;
  Real s, sp = 0, sm = 0;

  while ((u < ue) && (v < ve))
  { if (u->i < v->i) { u++; continue; }
    if (u->i > v->i) { v++; continue; }
    if ((s = u->x * v->x) >0) sp += s; else sm += s; u++, v++;
  }

  if (ausl(sp, -sm) < eps) return 0.; else return sp+sm;
}

void sp_Vector::operator*= (Real a)
{ for (Vec_el *u = A, *ue = u+n_A; u < ue; u->x *= a, u++); }
